package cmd

import (
	"github.com/golang-migrate/migrate/v4"
	"github.com/slicebit/qb"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/xbus/migration"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

type migrateLogger struct {
	log xbus.Logger
}

func (l migrateLogger) Printf(format string, v ...interface{}) {
	l.log.Noticef(format, v...)
}

func (migrateLogger) Verbose() bool {
	return false
}

var migrateCmd = cobra.Command{
	Use:   "migrate",
	Short: "Migrate the xbus database",
	Run: func(cmd *cobra.Command, args []string) {
		opts := xbus.LoggerOptions{Logtime: true, Debug: true, Trace: true}
		logger := xbus.GetLogger(&opts)

		_, db, err := model.NewDB("postgres",
			viper.GetString("database.dsn"),
			// Ignoring "database.max-connections". We need 2 connections, no
			// more, no _less_ (because the migrate driver will keep a connection
			// for itself)
			2,
			Logger, qb.LDefault)
		if err != nil {
			logger.Fatalf("Failed to initialize db access: %s", err)
		}
		defer db.Close()

		m, _, err := migration.New(db)
		if err != nil {
			Logger.Fatalf("Failed to init migration engine: %s", err)
		}
		defer m.Close()

		m.Log = migrateLogger{Logger}

		_, _, err = m.Version()
		if err == migrate.ErrNilVersion {
			// Check if some known tables are already here. If not, assume
			// an empty database and do a normal upgrade.
			row := db.Engine.DB().QueryRow(`SELECT EXISTS (
   SELECT 1
   FROM   pg_tables
   WHERE  schemaname = 'public'
   AND    tablename = 'envelope'
   ) `)
			var exists bool
			if err := row.Scan(&exists); err != nil {
				Logger.Fatalf("Error checking the db: %s", err)
			}
			// If some tables are already here, the database was created
			// with an old version of xbus and it needs to be upgraded manually
			// first
			if exists {
				Logger.Fatalf("The database is not versionned.\n" +
					"It probably means that it was created with an older version " +
					"of xbus.\n" +
					"Please apply migration scripts manually up to version " +
					"3.0.0-beta.14, then use the 'migrate' to finish the upgrades")
			}
		}

		if err := m.Up(); err != nil {
			if err == migrate.ErrNoChange {
				Logger.Noticef("The database is already up-to-date")
			} else {
				Logger.Fatalf("Failed to run migration: %s", err)
			}
		} else {
			Logger.Noticef("Database successfully upgraded")
		}
	},
}

func init() {
	RootCmd.AddCommand(&migrateCmd)
}
