package fullenv

import (
	"fmt"
	"os"
	"os/exec"
	"path"
	"strings"
	"syscall"
	"time"

	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	// Embed QB dialects we need
	_ "github.com/slicebit/qb/dialects/postgres"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

type clientSettings struct {
	name             string
	customXbusClient string
	autostart        bool
	inProcess        bool
	apikey           string
}

type clientRunner struct {
	dirname       string
	configFile    string
	xbusClientBin string
	logger        xbus.Logger
	settings      clientSettings
	cmd           *exec.Cmd
	inClient      *xbus.Client
	running       bool
}

func (c *clientRunner) configPath() string {
	return path.Join(c.dirname, c.configFile)
}

func (c *clientRunner) buildArgs(args ...string) (string, []string) {
	xbusclientCfgFilename := path.Join(
		c.dirname, c.configFile)

	args = append([]string{
		"--config", xbusclientCfgFilename,
	}, args...)

	return c.xbusClientBin, args
}

func (c *clientRunner) Settings() ProgramSettings {
	return ProgramSettings{
		c.settings.autostart,
		0,
	}
}

func (c *clientRunner) Running() bool {
	return c.running
}

func (c *clientRunner) Run(args ...string) error {
	program, finalArgs := c.buildArgs(args...)

	cmd := exec.Command(program, finalArgs...)
	cmd.Dir = c.dirname

	b, err := cmd.CombinedOutput()
	if err != nil {
		c.logger.Errorf("xbus-client: %s", b)
		return fmt.Errorf("Could not run xbus-client : %s", err)
	}

	return nil
}

func (c *clientRunner) Setup(xbusctl CmdRunner) error {
	if c.settings.inProcess {
		v := viper.New()
		xbus.DefineAndBindPFlags(v, &pflag.FlagSet{})
		v.SetConfigFile(c.dirname + "/" + c.configFile)
		err := v.ReadInConfig()
		if err != nil {
			return err
		}
		coptions := xbus.COptionsFromViper(
			v, api.Account_ACTOR, c.logger, c.dirname, c.configFile, true)
		client := xbus.NewClient(coptions, c.logger)
		if err := client.CreateKey(false); err != nil {
			return err
		}
		if err := client.CreateCSR("localhost", "", c.settings.name, false); err != nil {
			return err
		}
		if err := client.Connect(); err != nil {
			return err
		}
		status, err := client.Register()
		if err != nil {
			return err
		}
		if status != api.Registration_PENDING {
			return fmt.Errorf(
				"Invalid return from Register: %s (should be 'pending')", status)
		}
		if err := xbusctl.Run(
			"account", "accept", client.ID().String(),
		); err != nil {
			return err
		}

		for _, actor := range coptions.ActorList {
			if err := xbusctl.Run("actor", "accept", actor.Name); err != nil {
				return err
			}
		}

		status, err = client.Register()
		if err != nil {
			return err
		}
		if status != api.Registration_ACCEPTED {
			return fmt.Errorf(
				"Invalid return from Register: %s (should be 'accepted')",
				status)
		}
		// Reconnect so the new permissions are applied
		client.Close()
		if err := client.Connect(); err != nil {
			return err
		}
		c.inClient = client
	} else {
		if err := c.Run("register"); err != nil {
			return err
		}

		// Load the conf to read the ids
		v := viper.New()
		xbus.DefineAndBindPFlags(v, &pflag.FlagSet{})
		v.SetConfigFile(c.configPath())
		if err := v.ReadInConfig(); err != nil {
			return err
		}
		coptions := xbus.COptionsFromViper(
			v, api.Account_ACTOR, c.logger, c.dirname, c.configFile, true)

		// Accept the account and actor registration
		c.logger.Noticef("Accepting account %s", coptions.ID)
		if err := xbusctl.Run(
			"account", "accept", coptions.ID.String(),
		); err != nil {
			return err
		}

		for _, actor := range coptions.ActorList {
			c.logger.Noticef("Accepting %s", actor.ID)
			if err := xbusctl.Run(
				"actor", "accept", actor.ID.String(),
			); err != nil {
				return err
			}
		}

		if err := c.Run("register"); err != nil {
			return err
		}
	}
	if c.settings.apikey != "" {
		xbusctl.Run("account", "set-apikey", c.settings.name, c.settings.apikey)
	}
	return nil
}

func (c *clientRunner) Start() error {
	if c.running {
		return fmt.Errorf("Client '%s' already running", c.settings.name)
	}
	if c.settings.inProcess {
		if err := c.inClient.Startup(true); err != nil {
			return err
		}
	} else {
		program, args := c.buildArgs("serve")
		if c.settings.customXbusClient != "" {
			splitted := strings.Split(c.settings.customXbusClient, " ")
			program = splitted[0]
			args = append(splitted[1:], args[:len(args)-1]...)
		}
		cmd := exec.Command(program, args...)
		cmd.Dir = c.dirname

		if err := captureCmdOutputs("client-"+c.settings.name, cmd, c.logger); err != nil {
			return err
		}

		err := cmd.Start()
		if err != nil {
			return err
		}
		c.cmd = cmd
	}
	c.running = true
	return nil
}

func (c *clientRunner) Stop() error {
	if !c.running {
		return fmt.Errorf("Client '%s' is not running", c.settings.name)
	}
	if c.settings.inProcess {
		c.inClient.Shutdown()
		c.running = false
	} else {
		c.cmd.Process.Signal(os.Interrupt)
		var timer *time.Timer
		timer = time.AfterFunc(time.Second, func() {
			c.logger.Errorf(
				"Client '%s' still running. Attempting a SIGQUIT",
				c.settings.name)
			if err := c.cmd.Process.Signal(syscall.SIGQUIT); err != nil {
				c.logger.Errorf(
					"SIGQUIT failed with '%s', killing the process.", err)
				c.cmd.Process.Kill()
			} else {
				timer = time.AfterFunc(time.Second, func() {
					c.logger.Errorf(
						"Client '%s' still running. Killing it",
						c.settings.name)
					c.cmd.Process.Kill()
				})
			}
		})
		defer timer.Stop()
		return c.cmd.Wait()
	}

	return nil
}
