package http

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"net/url"
	"time"

	"github.com/gogo/protobuf/proto"
	"github.com/nats-rpc/gogo-nrpc"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// WebHookHandler ...
type WebHookHandler struct {
	url string
	sub *xbus.Subscription
}

// NewWebHookHandler ...
func NewWebHookHandler(actor *xbus.Actor) (*WebHookHandler, error) {
	url := actor.Options.GetConfigEntry("http.webhook.url")
	if url == "" {
		return nil, errors.New("No webhook on this actor")
	}
	handler := WebHookHandler{url: url}
	sub, err := actor.Subscribe(handler.Process)
	if err != nil {
		return nil, err
	}
	handler.sub = sub
	return &handler, nil
}

// Process a request
func (wh *WebHookHandler) Process(apc *xbus.ActorProcessingContext) error {
	// POST the request to the URL
	encoding := apc.Actor.Options.GetConfigEntry("http.webhook.content-type")
	if encoding == "" || encoding == "json" {
		encoding = "application/json"
	} else if encoding == "proto" || encoding == "protobuf" {
		encoding = "application/x-protobuf"
	}
	var payload []byte
	switch encoding {
	case "application/json":
		// TODO support item encodings (with 'application/json;base64')
		bytes, err := json.Marshal(apc.Request)
		if err != nil {
			return fmt.Errorf("Error encoding the payload: %s", err)
		}
		payload = bytes
	case "application/x-protobuf":
		bytes, err := proto.Marshal(apc.Request)
		if err != nil {
			return fmt.Errorf("Error encoding the payload %s", err)
		}
		payload = bytes
	default:
		return fmt.Errorf("Unsupported encoding: %s", encoding)
	}

	req, err := http.NewRequest("POST", wh.url, bytes.NewBuffer(payload))
	req = req.WithContext(apc.Context)
	req.Header.Set("Content-Type", encoding)

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		urlErr := err.(*url.Error)
		return &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: urlErr.Error(),
		}
	}
	var state api.ActorProcessingState
	if resp.StatusCode == http.StatusOK {
		dec := json.NewDecoder(resp.Body)
		if err := dec.Decode(&state); err != nil {
			return err
		}
	} else {
		state.Status = api.ActorProcessingState_ERROR
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = err.Error()
		}
		state.Messages = []api.LogMessage{
			{
				Time:  time.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			},
		}
	}
	if state.Status == api.ActorProcessingState_PENDING {
		apc.Detach()
	} else {
		apc.End(state.Status, state.Messages)
	}
	return nil
}

// Close everything
func (wh *WebHookHandler) Close() {
	wh.sub.Unsubscribe()
}
