// Copyright 2018 Orus.io. All rights reserved.

package sql

import (
	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

func toStorage(job model.Job) storage.Job {
	return storage.Job{
		JobID: storage.JobID{
			ProcessID: api.UUID(job.ProcessID),
			Node: storage.ProcessNode{
				ActorID: api.UUID(job.ActorID),
				NodeID:  job.NodeID,
			},
		},
		JobState: storage.JobState{
			Detached: job.Detached,
			Status:   job.Status,
		},
	}
}

func fromStorage(id storage.JobID, state storage.JobState) model.Job {
	return model.Job{
		ProcessID: uuid.UUID(id.ProcessID),
		ActorID:   uuid.UUID(id.Node.ActorID),
		NodeID:    id.Node.NodeID,
		Detached:  state.Detached,
		Status:    state.Status,
	}
}

// NewJobStorage initialize a new JobStorage
func NewJobStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *JobStorage {
	return &JobStorage{
		Base: Base{
			model:  model,
			db:     db,
			logger: logger,
		},
	}
}

// JobStorage implements storage.JobStorage
type JobStorage struct {
	Base
}

// NewJob ...
func (s *JobStorage) NewJob(
	id storage.JobID, state storage.JobState,
) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbJob = fromStorage(id, state)

	must(tx.Insert(&dbJob))
	tx.MustCommit()
}

// GetState returns the job state
func (s *JobStorage) GetState(id storage.JobID) storage.JobState {
	tx := s.MustBegin()
	defer tx.Rollback()

	var dbJob model.Job

	q := tx.Query(s.model.Job).
		Filter(
			s.model.Job.ProcessID.Eq(uuid.UUID(id.ProcessID)),
			s.model.Job.ActorID.Eq(uuid.UUID(id.Node.ActorID)),
			s.model.Job.NodeID.Eq(id.Node.NodeID),
		)
	must(q.One(&dbJob))

	return storage.JobState{
		Detached: dbJob.Detached,
		Status:   dbJob.Status,
	}
}

// SetState ...
func (s *JobStorage) SetState(id storage.JobID, state storage.JobState) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbJob model.Job

	q := tx.Query(s.model.Job).
		Filter(
			s.model.Job.ProcessID.Eq(uuid.UUID(id.ProcessID)),
			s.model.Job.ActorID.Eq(uuid.UUID(id.Node.ActorID)),
			s.model.Job.NodeID.Eq(id.Node.NodeID),
		)
	must(q.One(&dbJob))

	dbJob.Detached = state.Detached
	dbJob.Status = state.Status

	must(tx.Update(&dbJob, model.JobDetached, model.JobStatus))
	tx.MustCommit()
}

func (s *JobStorage) list(clauses ...qb.Clause) []storage.Job {
	tx := s.MustBegin()
	defer tx.Rollback()

	var dbJobs []model.Job

	q := tx.Query(s.model.Job)
	q = q.Filter(clauses...)
	q = q.OrderBy(s.model.Job.ID)

	must(q.All(&dbJobs))

	s.logger.Noticef("Loaded jobs: %v", dbJobs)

	var result = make([]storage.Job, len(dbJobs))

	i := 0
	for _, job := range dbJobs {
		result[i] = toStorage(job)
		i++
	}

	return result
}

// ListJobByStatus ...
func (s *JobStorage) ListJobByStatus(
	status api.ActorProcessingState_Status,
) []storage.Job {
	return s.list(s.model.Job.Status.Eq(status))
}

// ListPendingJobByActor returns the active jobs of a given actor
func (s *JobStorage) ListPendingJobByActor(
	actorID api.UUID, after *storage.JobID, limit int,
) []storage.Job {
	clauses := []qb.Clause{
		s.model.Job.ActorID.Eq(uuid.UUID(actorID)),
		s.model.Job.Status.In(
			api.ActorProcessingState_PENDING),
	}
	if after != nil {
		clauses = append(clauses,
			s.model.Job.ID.Gt(
				s.model.Job.GetMapper().Table().Select(
					s.model.Job.ID,
				).
					Where(
						s.model.Job.ProcessID.Eq(after.ProcessID),
						s.model.Job.ActorID.Eq(after.Node.ActorID),
						s.model.Job.NodeID.Eq(after.Node.NodeID),
					),
			),
		)
	}
	list := s.list(clauses...)
	// TODO add the limit in the SQL query
	if len(list) > limit {
		list = list[:limit]
	}
	return list
}

// ListJobByProcess returns the active jobs of a given process
func (s *JobStorage) ListJobByProcess(processID api.UUID) []storage.Job {
	return s.list(
		s.model.Job.ProcessID.Eq(uuid.UUID(processID)),
		s.model.Job.Status.In(
			api.ActorProcessingState_PENDING,
			api.ActorProcessingState_RUNNING))
}

// Purge removes jobs
func (s *JobStorage) Purge(processID api.UUID) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	_, err := tx.GetEngine().Exec(
		s.model.Job.GetMapper().Table().Delete().
			Where(s.model.Job.ProcessID.Eq(uuid.UUID(processID))))
	if err == nil {
		tx.MustCommit()
	}
	return err
}
