// Copyright 2016 Orus.io. All rights reserved.

package tools

import (
	"encoding/json"
	"errors"
	"fmt"

	api "bitbucket.org/orus-io/go-xbus/api"
)

var (
	// ErrWrongEnvelope is returned by validator if an envelope with a wrong
	// ID is passed to it.
	ErrWrongEnvelope = errors.New("Wrong envelope")
)

// EnvelopeValidator handle a stream of chunks of envelope and
// calculate checksum on the fly.
// The event chunks in the stream must be ordered.
// Its state can be easily saved to later continuation
type EnvelopeValidator struct {
	EnvelopeID      api.UUID
	ExpectedEvents  []api.UUID
	EventTypesKnown bool
	Events          map[api.UUID]EventValidator // Index if the last chunk of each event
	Complete        bool
	Valid           bool
}

// EventValidator handles a stream of event chunks of the same event
type EventValidator struct {
	EventID          api.UUID
	Type             string
	ItemCount        uint64
	Checksum         api.Checksum
	CurrentIndex     uint64
	CurrentItemCount uint64
	CurrentChecksum  api.Checksum
	Complete         bool
	Valid            bool
	Error            error
}

// NewEnvelopeValidator returns a new validator, and optionnaly feeds it
// some envelopes
func NewEnvelopeValidator() EnvelopeValidator {
	return EnvelopeValidator{
		Events: make(map[api.UUID]EventValidator),
	}
}

// IsNext returns true if the given fragment can be next in line,
// ie its events index are just after the ones already validated
func (v *EnvelopeValidator) IsNext(envelope api.Envelope) bool {
	for _, event := range envelope.Events {
		ev := v.Events[event.ID]
		if ev.CurrentIndex+1 != event.Index {
			return false
		}
	}
	return true
}

// Add adds an envelope
func (v *EnvelopeValidator) Add(envelope api.Envelope) error {
	if v.EnvelopeID.IsZero() {
		v.EnvelopeID = envelope.ID
	}
	if envelope.ID != v.EnvelopeID {
		return ErrWrongEnvelope
	}
	if len(envelope.EventIDs) != 0 {
		v.ExpectedEvents = envelope.EventIDs
	}
	var err error
	for _, event := range envelope.Events {
		ev := v.Events[event.ID]
		ev, err = ev.Update(event)
		if err != nil {
			return err
		}
		if ev.Error != nil {
			return ev.Error
		}
		v.Events[event.ID] = ev
	}
	v.update()
	if v.Complete {
		return v.Check()
	}
	return nil
}

func (v *EnvelopeValidator) update() {
	if !v.EventTypesKnown &&
		len(v.ExpectedEvents) == len(v.Events) {
		var foundUnkownType bool
		for _, ev := range v.Events {
			if ev.Type == "" {
				foundUnkownType = true
				break
			}
		}
		if !foundUnkownType {
			v.EventTypesKnown = true
		}
	}
	var complete, valid = true, true
	for _, id := range v.ExpectedEvents {
		ev, ok := v.Events[id]
		if !ok {
			complete = false
			valid = false
		} else if !ev.Complete {
			complete = false
			valid = false
		} else if !ev.Valid {
			valid = false
		}
		if !complete {
			break
		}
	}
	v.Complete = complete
	v.Valid = valid
}

// Check compare the current validation state with a final state
func (v *EnvelopeValidator) Check() error {
	if len(v.ExpectedEvents) != len(v.Events) {
		return fmt.Errorf(
			"Expecting %d events, got %d",
			len(v.ExpectedEvents), len(v.Events))
	}
	if !v.Valid {
		if v.Complete {
			return fmt.Errorf("Checksum error")
		}
		return fmt.Errorf("Incomplete error")
	}
	return nil
}

// LoadEventValidator loads a EventValidator from a string
func LoadEventValidator(s string) (ev EventValidator, err error) {
	err = json.Unmarshal([]byte(s), &ev)
	return
}

// DumpEventValidator store a EventValidator in a string
func DumpEventValidator(ev EventValidator) (string, error) {
	b, err := json.Marshal(ev)
	if err != nil {
		return "", err
	}
	return string(b), nil
}

// NewEventValidator initialize a EventValidator
func NewEventValidator(event api.Event) (EventValidator, error) {
	return EventValidator{}.Update(event)
}

// Update calculates a new intermediary state
func (v EventValidator) Update(event api.Event) (EventValidator, error) {
	if v.CurrentIndex == 0 {
		// This is the fist event we see, a few initializations
		// are required
		if event.ID.IsZero() {
			return EventValidator{}, fmt.Errorf("First Event must have an ID")
		}
		if event.Type == "" {
			return EventValidator{}, fmt.Errorf("First Event must have a Type")
		}
		v.EventID = event.ID
		v.Type = event.Type
		v.CurrentChecksum = api.Checksum(0).
			Update(event.ID[:16]).
			Update([]byte(event.Type))
	}
	if event.Index != v.CurrentIndex+1 {
		return EventValidator{}, fmt.Errorf(
			"Invalid event %s index. Expected %d, got %d",
			event.ID, v.CurrentIndex+1, event.Index)
	}
	if v.Complete && len(event.Items) != 0 {
		return EventValidator{}, fmt.Errorf(
			"Event validator already complete. Next events can only contains " +
				"ItemCount & Checksum, not Items")
	}
	if event.Type != "" && v.Type != event.Type {
		return EventValidator{}, fmt.Errorf(
			"Event type changed. Previously had %s, and got %s at index %d",
			v.Type, event.Type, event.Index)
	}
	if event.ItemCount != 0 {
		if v.ItemCount == 0 {
			v.ItemCount = event.ItemCount
		} else if event.ItemCount != v.ItemCount {
			return EventValidator{}, fmt.Errorf(
				"Inconsistent ItemCount at index %d. Previously had %d, got %d",
				event.Index, v.ItemCount, event.ItemCount,
			)
		}
	}
	if event.Checksum != 0 {
		if v.Checksum == 0 {
			v.Checksum = event.Checksum
		} else if v.Checksum != event.Checksum {
			return EventValidator{}, fmt.Errorf(
				"Inconsistent Checksum at index %d. Previously had 0x%x, got 0x%x",
				event.Index, v.Checksum, event.Checksum,
			)
		}
	}

	v.CurrentIndex++
	v.CurrentItemCount += uint64(len(event.Items))
	for _, item := range event.Items {
		v.CurrentChecksum = v.CurrentChecksum.Update(item)
	}
	if v.ItemCount != 0 && v.ItemCount == v.CurrentItemCount {
		v.Complete = true
	}
	if v.ItemCount == 0 && v.ItemCount == v.CurrentItemCount && v.Checksum != 0 && v.Checksum == v.CurrentChecksum {
		v.Complete = true
		v.Valid = true
	}
	if v.Complete {
		if v.Checksum == v.CurrentChecksum {
			v.Valid = true
		} else {
			v.Error = fmt.Errorf(
				"Checksum error. Expected 0x%x, got 0x%x", v.Checksum, v.CurrentChecksum)
		}
	}
	return v, nil
}
