package xbus

import (
	"bytes"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"io/ioutil"
	"os"
	"path"
	"strings"
	"time"

	"github.com/nats-io/go-nats"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	"gopkg.in/yaml.v2"

	api "bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

const (
	confKeyAccountID   = "account-id"
	confKeyAccountName = "account-name"
	confKeyNatsURL     = "nats-url"
	confKeyNatsTimeout = "nats-timeout"
	confKeyForceStart  = "force-start"

	confKeyPersistentStore = "persistent-store"

	confKeyPrivateKey  = "tls.private-key"
	confKeyCertificate = "tls.certificate"
	confKeyCSR         = "tls.csr"
	confKeyRootCA      = "tls.rootca"
	confKeyTLSTimeout  = "tls.timeout"

	confKeySentryDSN = "sentry.dsn"
)

func fileExists(path string) bool {
	if _, err := os.Stat(path); err != nil {
		if os.IsNotExist(err) {
			return false
		}
		panic(err)
	}
	return true
}

func readPEM(blockType string, pemOrPath string) []byte {
	var (
		bytes []byte
		err   error
	)
	if strings.Contains(pemOrPath, "-----BEGIN") {
		if !strings.Contains(pemOrPath, "-----BEGIN "+blockType+"-----") {
			panic(fmt.Sprintf(
				"Invalid pem block type. Expected %s, got:\n%s",
				blockType, pemOrPath))
		}
		bytes = []byte(pemOrPath)
	} else if fileExists(pemOrPath) {
		bytes, err = ioutil.ReadFile(pemOrPath)
		if err != nil {
			panic(err)
		}
	} else {
		panic("Not a file nor a PEM block: " + pemOrPath)
	}
	block, _ := pem.Decode(bytes)
	if block == nil {
		panic("Could not decode PEM data")
	}
	return block.Bytes
}

// COptionsFromViper prepare a COptions from a Viper conf
// If writeback is true and the viper backend allows it, the
// option changes are wrote back to the configuration backend (
// most likely a configuration file).
func COptionsFromViper(
	v *viper.Viper,
	accountType api.Account_Type,
	logger Logger,
	configDir string,
	configName string,
	writeback bool) COptions {
	opts := COptions{
		Name:        v.GetString(confKeyAccountName),
		Type:        accountType,
		ForceStart:  v.GetBool(confKeyForceStart),
		NatsTimeout: v.GetDuration(confKeyNatsTimeout),
		ConnOptions: nats.Options{
			Url: v.GetString(confKeyNatsURL),
			Timeout: time.Duration(
				v.GetFloat64(confKeyTLSTimeout) * float64(time.Second)),
			AsyncErrorCB: func(c *nats.Conn, s *nats.Subscription, err error) {
				if s != nil {
					logger.Errorf("Error on sub %s: %s", s.Subject, err)
				} else {
					logger.Errorf("Error %s", err)
				}
			},
		},
	}

	if s := v.GetString(confKeyAccountID); s != "" {
		opts.ID = api.MustUUIDFromString(s)
	}

	for i := 0; ; i++ {
		prefix := fmt.Sprintf("actors.%d.", i)
		name := v.GetString(prefix + "name")
		if name == "" {
			break
		}
		var id api.UUID
		if s := v.GetString(prefix + "id"); s != "" {
			id = api.MustUUIDFromString(s)
		}
		var aKind api.Actor_Kind
		if s := v.GetString(prefix + "kind"); s != "" {
			if err := aKind.UnmarshalText([]byte(s)); err != nil {
				panic(err)
			}
		}
		var aStatus api.Actor_Status
		if s := v.GetString(prefix + "status"); s != "" {
			if err := aStatus.UnmarshalText([]byte(s)); err != nil {
				panic(err)
			}
		}
		roles := v.GetStringSlice(prefix + "roles")
		serviceType := v.GetString(prefix + "service.type")
		serviceSettings := v.GetStringMap(prefix + "service.settings")

		opts.ActorList = append(opts.ActorList, ActorOptions{
			Actor: api.Actor{
				ID:     id,
				Name:   name,
				Kind:   aKind,
				Status: aStatus,
				Roles:  roles,
			},
			Service: ServiceOptions{
				Type:     serviceType,
				Settings: serviceSettings,
			},
		})
	}

	if s := v.GetString(confKeyPrivateKey); s != "" {
		logger.Debugf("Loading pkey")
		key, err := x509.ParseECPrivateKey(readPEM("EC PRIVATE KEY", s))
		if err != nil {
			panic(err)
		}
		opts.PrivKey = key
	}

	if s := v.GetString(confKeyCSR); s != "" {
		logger.Debugf("Loading CSR")
		opts.CSR = readPEM("CERTIFICATE REQUEST", s)
	}

	if s := v.GetString(confKeyCertificate); s != "" {
		logger.Debugf("Loading certificate")
		opts.Certificate = readPEM("CERTIFICATE", s)
	}

	if s := v.GetString(confKeyRootCA); s != "" {
		logger.Debugf("Loading RootCA")

		opts.RootCAs = [][]byte{
			readPEM("CERTIFICATE", s),
		}
	}

	if s := v.GetString(confKeySentryDSN); s != "" {
		opts.SentryDSN = s
	}

	{
		var ckps string
		if ckps = v.GetString(confKeyPersistentStore); ckps == "" {
			ckps = configDir
		}
		store, err := NewFilePersistentStore(path.Join(ckps, "pstore.db"))
		if err != nil {
			panic(err)
		}
		opts.PersistentStore = store
	}

	if writeback {
		cfgFile := v.ConfigFileUsed()
		if cfgFile == "" {
			cfgFile = path.Join(configDir, configName+".yaml")
		}
		opts.ChangeCallback = func(opts COptions) {
			confMap := v.AllSettings()
			confMap[confKeyAccountID] = opts.ID
			confMap[confKeyAccountName] = opts.Name
			confMap[confKeyNatsURL] = opts.ConnOptions.Url
			if opts.NatsTimeout != 0 {
				confMap[confKeyNatsTimeout] = opts.NatsTimeout
			}
			if opts.ConnOptions.Timeout != 4*time.Second {
				confMap[confKeyTLSTimeout] = float64(opts.ConnOptions.Timeout) / float64(time.Second)
			}

			if opts.PrivKey != nil {
				if s := v.GetString(confKeyPrivateKey); s == "" {
					b, err := tools.PKeyToPEM(opts.PrivKey)
					if err != nil {
						panic(err)
					}
					confMap[confKeyPrivateKey] = string(b)
				} else {
					confMap[confKeyPrivateKey] = s
				}
			}
			if opts.CSR != nil {
				if s := v.GetString(confKeyCSR); s == "" {
					b := pem.EncodeToMemory(&pem.Block{
						Type:  "CERTIFICATE REQUEST",
						Bytes: opts.CSR,
					})
					confMap[confKeyCSR] = string(b)
				} else {
					confMap[confKeyCSR] = s
				}
			}
			if opts.Certificate != nil {
				b := pem.EncodeToMemory(&pem.Block{
					Type:  "CERTIFICATE",
					Bytes: opts.Certificate,
				})
				confMap[confKeyCertificate] = string(b)
			}
			if opts.RootCAs != nil {
				var buf bytes.Buffer
				for _, cert := range opts.RootCAs {
					pem.Encode(&buf, &pem.Block{
						Type:  "CERTIFICATE",
						Bytes: cert,
					})
				}
				confMap[confKeyRootCA] = buf.String()
			} else if s := v.GetString(confKeyRootCA); s != "" {
				confMap[confKeyRootCA] = s
			}
			if len(opts.ActorList) != 0 {
				actors := map[string]ActorOptions{}
				for i, actor := range opts.ActorList {
					actors[fmt.Sprintf("%d", i)] = actor
				}
				confMap["actors"] = actors
			}
			if opts.SentryDSN != "" {
				confMap["sentry.dsn"] = opts.SentryDSN
			}
			b, err := yaml.Marshal(confMap)
			if err != nil {
				panic(err)
			}
			logger.Noticef("Saving configuration to: %s", cfgFile)
			dirName := path.Dir(cfgFile)
			if _, err = os.Stat(dirName); err != nil {
				if os.IsNotExist(err) {
					logger.Debugf("Create %s", dirName)
					err = os.MkdirAll(dirName, 0700)
				}
				if err != nil {
					panic(err)
				}
			}
			if err := ioutil.WriteFile(cfgFile, b, 0600); err != nil {
				panic(err)
			}
		}
	}
	return opts
}

// DefineAndBindPFlags populates a viper.Viper and pflags.FlagSet
// with all the options suitable for a xbus client
func DefineAndBindPFlags(v *viper.Viper, flags *pflag.FlagSet) {
	flags.StringP(confKeyNatsURL, "u", "nats://localhost:4222", "NATS url")
	v.BindPFlag(confKeyNatsURL, flags.Lookup(confKeyNatsURL))

	flags.Duration(confKeyNatsTimeout, 0, "NATS Timeout")
	v.BindPFlag(confKeyNatsTimeout, flags.Lookup(confKeyNatsTimeout))

	flags.String(confKeyAccountID, "", "Account ID")
	v.BindPFlag(confKeyAccountID, flags.Lookup(confKeyAccountID))

	flags.String(confKeyAccountName, "", "Account Name")
	v.BindPFlag(confKeyAccountName, flags.Lookup(confKeyAccountName))

	flags.String(confKeyPrivateKey, "", "Private key (a path or a PEM encoded key)")
	v.BindPFlag(confKeyPrivateKey, flags.Lookup(confKeyPrivateKey))

	flags.String(confKeyCertificate, "", "Certificate (a path or a PEM encoded cert)")
	v.BindPFlag(confKeyCertificate, flags.Lookup(confKeyCertificate))

	flags.String(confKeyCSR, "", "CSR (a path or a PEM encoded CSR)")
	v.BindPFlag(confKeyCSR, flags.Lookup(confKeyCSR))

	flags.Float64(confKeyTLSTimeout, 4.0, "TLS Timeout")
	v.BindPFlag(confKeyTLSTimeout, flags.Lookup(confKeyTLSTimeout))

	flags.Bool(
		confKeyForceStart, false,
		"Enable force start (stay up even if the server is not reachable on start)")
	v.BindPFlag(confKeyForceStart, flags.Lookup(confKeyForceStart))

	flags.String(confKeySentryDSN, "", "Sentry DSN")
	v.BindPFlag(confKeySentryDSN, flags.Lookup(confKeySentryDSN))
}

// COptionsFromFile load a client configuration from a file
func COptionsFromFile(path string, accountType api.Account_Type, logger Logger) (COptions, error) {
	in, err := os.Open(path)
	if err != nil {
		return COptions{}, err
	}

	xbusViper := viper.New()
	xbusFlags := &pflag.FlagSet{}

	DefineAndBindPFlags(xbusViper, xbusFlags)

	xbusViper.SetConfigType("yaml")
	xbusViper.ReadConfig(in)

	options := COptionsFromViper(
		xbusViper, accountType, logger, "", "", false,
	)
	return options, nil
}

// ClientFromFile loads a client from a configuration file
func ClientFromFile(path string, accountType api.Account_Type, logger Logger) (*Client, error) {
	options, err := COptionsFromFile(path, accountType, logger)

	if err != nil {
		return nil, err
	}

	client := NewClient(options, logger)
	return client, nil
}
